{#
    basic/globaltoc.html
    ~~~~~~~~~~~~~~~~~~~~

    Sphinx sidebar template: global table of contents.

    :copyright: Copyright 2007-2011 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
#}
<h3>{{ _('Manual') }}</h3>
{{ toctree() }}
<a href="{{ pathto(master_doc) }}">Back to Welcome</a>
